/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "count.h"
#include "dict.h"
#include "vector.h"

int hash_size(int size);

extern int debug;

count new_count(int n, int d) {
  count ct;

  ct = malloc(sizeof(*ct));

  ct->n = n;
  ct->d = d;
  ct->size = hash_size(DEFAULT_SIZE);
  ct->nelem = 0;

  ct->t = calloc(ct->size, sizeof(struct count_hash_node *));

  return ct;

}

int count_read(count ct, const char *filename) {
  FILE *fp;
  int k;
  double *c;
  static int ks[MAX_DEPTH];

  fp = fopen(filename, "rb");

  k = 0;
  c = malloc(sizeof(double) * ct->d);

  fread(ks, sizeof(int), ct->n, fp);
  fread(c, sizeof(double), ct->d, fp);

  while (!feof (fp)) {

    k++;
    if (k%10000 == 0) {
      fprintf(stderr, "%d items processed.\n", k);
    }

    count_insert(ct, ks, c);

    fread(ks, sizeof(int), ct->n, fp);
    fread(c, sizeof(double), ct->d, fp);
  }

  free(c);
  fclose(fp);

  return 0;

}

int count_write(count ct, const char *filename) {
  FILE *fp;
  int i, j, skip;
  struct count_hash_node *cur;

  fp = fopen(filename, "wb");

  for (i = 0; i < ct->size; i++)
    for (cur = ct->t[i]; cur != NULL; cur = cur->next) {
      fwrite(cur->ks, sizeof(int), ct->n, fp);
      fwrite(&cur->c, sizeof(double), ct->d, fp);
    }

  fclose(fp);

  return 0;

}

void count_delete(count ct) {
  if (ct) {
    free (ct->t);
    free (ct);
  }
}


int count_rehash(count ct, int new_size) {
  struct count_hash_node **old_t, *cur, *next;
  int old_size;
  int i;

  ct->nelem = 0;

  old_t = ct->t;
  old_size = ct->size;

  ct->size = hash_size(new_size);
  ct->t = calloc(ct->size, sizeof(struct count_hash_node *));
  if (ct->t == NULL) {
    fprintf(stderr, "Rehashing to %d failed.\n", ct->size);
    ct->t = old_t; ct->size = old_size;
    return -1;
  }

  fprintf(stderr, "Rehashing to %d...\n", ct->size);

  for (i=0; i<old_size; i++) {
    cur = old_t[i];
    while (cur != NULL) {
      count_insert(ct, cur->ks, &cur->c);
      next = cur->next;
      free(cur->ks);
      free(cur);
      cur = next;
    }
  }

  free(old_t);

  return 0;
}

/* If ks was found, returns a pointer to a pointer to the node.
   If ks was not found, returns a pointer to the null pointer where
     the node should be inserted. */

struct count_hash_node ** count_get_prehashed(count ct, int ks[], int hash) {
  struct count_hash_node **pcur;

  pcur = &ct->t[hash];
      
  while (*pcur != NULL) {
    if (!vector_compare(ks, (*pcur)->ks, ct->n))
      return pcur;
    pcur = &(*pcur)->next;
  }

  return pcur;
}

struct count_hash_node ** count_get(count ct, int ks[]) {
  int hash;
  hash = vector_hash(ks, ct->n, ct->size);
  return count_get_prehashed(ct, ks, hash);
}

int count_insert(count ct, int ks[], double c[]) {

  struct count_hash_node **pcur, *cur;

  pcur = count_get(ct, ks);
  cur = *pcur;
  
  if (cur == NULL) {
    cur = *pcur = malloc(sizeof(struct count_hash_node)+(ct->d-1)*sizeof(double));
    if (cur == NULL)
      return -1;

    cur->ks = malloc(sizeof(int) * ct->n);
    if (cur->ks == NULL)
      return -1;
    memcpy(cur->ks, ks, sizeof(int) * ct->n);
    memcpy(&cur->c, c, sizeof(double) * ct->d);

    cur->next = NULL;

    ct->nelem++;

    if (ct->nelem > LOAD_FACTOR * ct->size) {
      count_rehash(ct, ct->size * 2);
    }
  } else {
    memcpy(&cur->c, c, sizeof(double) * ct->d);
  }
  
  return 0;
}

int count_count(count ct, int ks[], int i, double delta, double *ptotal) {
  struct count_hash_node **pcur, *cur;
  int j, retval;

  pcur = count_get(ct, ks);
  cur = *pcur;
  
  if (cur == NULL) {
    cur = *pcur = malloc(sizeof(struct count_hash_node)+(ct->d-1)*sizeof(double));
    if (cur == NULL)
      return -1;

    cur->ks = malloc(sizeof(int) * ct->n);
    if (cur->ks == NULL)
      return -1;
    memcpy(cur->ks, ks, sizeof(int) * ct->n);
    for (j=0; j<ct->d; j++)
      (&cur->c)[j] = 0.0;
    (&cur->c)[i] = delta;

    cur->next = NULL;

    ct->nelem++;

    if (ct->nelem > LOAD_FACTOR * ct->size) {
      count_rehash(ct, ct->size * 2);
    }

    retval = 1;

  } else {
    (&cur->c)[i] += delta;

    retval = 0;
  }
  
  if (ptotal)
    *ptotal = (&cur->c)[i];

  return retval;
}

double count_lookup(count ct, int ks[], int i) {
  double *c;

  c = count_mlookup(ct, ks);

  if (c == NULL)
    return 0.0;
  else
    return c[i];
}

double * count_mlookup(count ct, int ks[]) {
  struct count_hash_node *cur;
  cur = *count_get(ct, ks);
  if (cur == NULL) {
    return NULL;
  } else {
    return &cur->c;
  }
}

#if 0

dict d;

int main (int argc, char *argv[]) {
  count ct;
  FILE *fp;
  float p, lambda;
  int ks[8];

  d = new_dict(20000);

  ct = new_count(8, 1000000);
  fp = fopen("/nldb5/dchiang/hybrid/tig0/tig0.dump.0.5", "r");
  count_read(ct, fp);


  ks[0] = dict_lookup(d, "NNS");
  ks[1] = dict_lookup(d, "521");
  ks[2] = dict_lookup(d, "MOD");
  ks[3] = dict_lookup(d, "246");
  ks[4] = dict_lookup(d, "11");
  ks[5] = dict_lookup(d, "0");
  ks[6] = dict_lookup(d, "IN");
  ks[7] = dict_lookup(d, "FOR");

  count_lookup(ct, ks, &p, &lambda);
  printf ("%f %f\n", p, lambda);
}

#endif
