/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dict.h"

dict new_dict(int m) {
  dict retval;

  retval = malloc(sizeof (*retval));
  retval->m = m;
  retval->t = calloc(m, sizeof(const char *));

  return retval;
}

void dict_delete(dict d) {
  free(d->t);
  free(d);
}

/* Double hashing. Knuth 6.4, algorithm D */

int hash1(const char *s, int m) {
  unsigned int n = 0;

  while (*s) {
    n = ((n << 7) + *s) % m;
    s++;
  }

  return (int)n;
}

int hash2(const char *s, int m) {
  unsigned int n = 0;

  while (*s) {
    n = ((n << 7) + *s) % (m-2);
    s++;
  }

  return (int)(n+1);
}

dict read_dict(FILE *fp) {
  static char buf[MAX_LENGTH];
  int i, l, m;
  dict d;

  fgets(buf, sizeof(buf), fp);
  sscanf(buf, "%d", &m);
  d = new_dict(m);

  for (i=0; i<d->m; i++) {
    fgets(buf, sizeof(buf), fp);
    l = strlen(buf);
    if (l>0 && buf[l-1] == '\n')
      buf[l-1] = '\0';
    if (buf[0])
      d->t[i] = strdup(buf);
    else
      d->t[i] = NULL;
  }

  return d;
}

int dict_write(dict d, FILE *fp) {
  int i;

  fprintf(fp, "%d\n", d->m);

  for (i=0; i<d->m; i++) {
    if (d->t[i])
      fputs(d->t[i], fp);
    fputc('\n', fp);
  }

  return 0;
}

int dict_insert(dict d, const char *k) {
  int h, h1, h2;
  
  h = h1 = hash1(k,d->m); h2 = hash2(k,d->m);

  while (d->t[h] && strcmp(k, d->t[h])) {
    h = (h + h2) % d->m;
    if (h == h1) {
      fprintf(stderr, "dict_insert(): hash table full\n");
      return -1; /* Hash table full */
    }
  }

  if (!d->t[h])
    d->t[h] = strdup(k);

/*  printf ("%s -> %d\n", k, h);*/

  return h;
}

int dict_lookup(dict d, const char *k) {
  int h, h1, h2;
  
  h = h1 = hash1(k,d->m); h2 = hash2(k,d->m);

  while (d->t[h] && strcmp(k, d->t[h])) {
    h = (h + h2) % d->m;
    if (h == h1)
      return -1; /* not found */
  }

  if (d->t[h])
    return h;
  else
    return -1;

}

int dict_rlookup(dict d, int i, char *buf, int buf_size) {
  if (i<0 || d->t[i] == NULL) {
    strncpy(buf, "?", buf_size);
    return -1;
  }

  strncpy(buf, d->t[i], buf_size);
  if (strlen(d->t[i]) > buf_size-1) {
    buf[buf_size-1] = '\0';
    return buf_size-1;
  } else {
    return strlen(d->t[i]);
  }
}


#if 0

int main (int argc, char *argv[]) {
  char buf[MAX_LENGTH];
  dict d = new_dict(1499);
  int i, h, l;
  FILE *fp;
  
  fp = fopen(argv[1], "r");

  while (!feof(fp)) {
    fgets(buf, sizeof(buf), fp);
    l = strlen(buf);
    if (l>0 && buf[l-1] == '\n')
      buf[l-1] = '\0';
    if (buf[0])
      dict_insert(d, buf);
  }

  fclose(fp);

  dict_write(d, "/tmp/dict.test");
  dict_read(d, "/tmp/dict.test"); /* massive memory leak */

  for (i=0; i<1499; i++) {
    if (d->t[i]) {
      h = dict_lookup(d, d->t[i]);
      if (h < 0)
	fprintf(stderr, "%s not found\n", d->t[i]);
      else if (h != i)
	fprintf(stderr, "dict_lookup(\"%s\"): expected %d, found %d\n", d->t[i], i, h);
    }
  }
}

#endif




