/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <config.h>

#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "count.h"
#include "vector.h"

#include "prob.h"

extern dict d;

/* Global options */

float C = 5.0;
int DIVERSITY_CUTOFF = 0.0;
int debug = 0;
int rewrite_lambdas = 1;

/* BUGS: prob_count() shouldn't be used with a cache */

prob new_prob(int m, int n, int *flags, char **names) {
  prob retval;
  int i;

  retval = malloc(sizeof(*retval));

  retval->m = m;
  retval->n = n;

  retval->nct = malloc((n-m)*sizeof(count));
  retval->udct = malloc((n-m)*sizeof(count));
  retval->uct = malloc((n-m)*sizeof(count));

  for (i=0; i<n-m; i++)
    if (flags[i]) {
      retval->udct[i] = new_count(i+1, 2);
      retval->uct[i] = new_count(i+1, 1);
      retval->nct[i] = new_count(i+m+1, 1);
    } else {
      retval->udct[i] = retval->uct[i] = retval->nct[i] = NULL;
    }

  for (i=0; i<n-m; i++) {
    retval->flags[i] = flags[i];
  }

  for (i=0; i<n; i++) {
    if (names && names[i])
      retval->names[i] = strdup(names[i]);
  }


  retval->cache_size = 0;
  retval->cache_p = NULL;
  retval->cache_ks = NULL;
  retval->hits = retval->lookups = 0;

  retval->next = NULL;

  return retval;
}

int prob_set_cache(prob p, int cache_size) {
  p->cache_size = cache_size;
  p->cache_p = malloc(cache_size * sizeof(double));
  p->cache_ks = calloc(cache_size, sizeof(int *));
  if (p->cache_p == NULL || p->cache_ks == NULL)
    return -1;

  p->hits = p->lookups = 0;
  return 0;
}

int prob_read(prob p, const char *prefix) {
  char filename[100];

  int i;

  for (i=0; i<p->n-p->m; i++)
    if (p->flags[i]) {

      sprintf(filename, "%s.%d.udct", prefix, i);
      count_read(p->udct[i], filename);

      sprintf(filename, "%s.%d.uct", prefix, i);
      count_read(p->uct[i], filename);

      sprintf(filename, "%s.%d.nct", prefix, i);
      count_read(p->nct[i], filename);

    }

  return 0;
}

int prob_read_lambdas(prob p, const char *prefix) {
  char filename[100];

  int i;

  for (i=0; i<p->n-p->m; i++)
    if (p->flags[i]) {
      sprintf(filename, "%s.%d.udct", prefix, i);
      count_read(p->udct[i], filename);
    }

  return 0;
}

int prob_write(prob p, const char *prefix) {
  int i;
  char filename[100];

  for (i=0; i<p->n-p->m; i++)
    if (p->flags[i]) {

      sprintf(filename, "%s.%d.udct", prefix, i);
      count_write(p->udct[i], filename);

      sprintf(filename, "%s.%d.uct", prefix, i);
      count_write(p->uct[i], filename);

      sprintf(filename, "%s.%d.nct", prefix, i);
      count_write(p->nct[i], filename);

    }
  return 0;
}

void prob_delete(prob p) {
  int i;
  
  if (p) {
    if (p->cache_size > 0) {
      free(p->cache_p);
      free(p->cache_ks);
    }

    for (i=0; i<p->n-p->m; i++) {
      count_delete(p->nct[i]);
      count_delete(p->udct[i]);
      count_delete(p->uct[i]);
    }

    free(p);
  }
}

int prob_count(prob p, int *ks, double delta) {
  int i;
  double total;

  for (i=0; i<p->n-p->m; i++)
    if (p->flags[i]) {
      if (!rewrite_lambdas) {
	if (count_lookup(p->udct[i], ks+p->m, 0) == 0.0)
	  continue;
      } else {
	if (count_count(p->udct[i], ks+p->m, 0, delta, NULL) < 0)
	  return -1;
      }
      if (count_count(p->uct[i], ks+p->m, 0, delta, NULL) < 0)
	return -1;
      if (count_count(p->nct[i], ks, 0, delta, &total) < 0)
	return -1;

      if (rewrite_lambdas && total > DIVERSITY_CUTOFF && total-delta <= DIVERSITY_CUTOFF) {
	if (count_count(p->udct[i], ks+p->m, 1, 1.0, NULL) < 0)
	  return -1;
      }
    }
  return 0;
}

double prob_lookup(prob p, int *ks) {
  double retval, lambda;
  int hash;
  int i, j, oldi;
  double excess;
  double e;
  static double ncounts[MAX_DEPTH], ucounts[MAX_DEPTH];
  static double u0counts[MAX_DEPTH], d0counts[MAX_DEPTH]; /* for lambdas */
  double *c;
  char buf[100];

  p->lookups++;

  if (debug) {
    for (i=0; i<p->n; i++) {
      dict_rlookup(d, ks[i], buf, sizeof(buf));
      printf (" %s", buf);
    }
  }


  /* First see if it's in the cache */
  hash = vector_hash(ks, p->n, p->cache_size);

  if (p->cache_ks && p->cache_ks[hash] && !vector_compare(ks, p->cache_ks[hash], p->n)) {
    p->hits++;

    if (debug) {
      printf (" => %f (cached)\n", p->cache_p[hash]);
    }
    return p->cache_p[hash];
  }

  /* Lookup counts */

  for (i=0; i<p->n-p->m; i++) {
    if (p->flags[i]) {
      c = count_mlookup(p->udct[i], ks+p->m);
      if (c) {
	u0counts[i] = c[0];
	d0counts[i] = c[1];
      } else {
	u0counts[i] = 0.0;
	d0counts[i] = 0.0;
      }

      ucounts[i] = count_lookup(p->uct[i], ks+p->m, 0);

      if (ucounts[i] < SMALL) {
	for (j=i+1; j<p->n-p->m; j++) {
	  ucounts[j] = u0counts[j] = 0.0;
	  d0counts[j] = 0.0;
	}
	break;
      }

      ncounts[i] = count_lookup(p->nct[i], ks, 0);
    }
  }

  if (debug > 1) {
    printf("\nn ");
    for (i=0; i<p->n-p->m; i++)
      if (p->flags[i])
	printf("%11.2f", ncounts[i]);
    printf("\nu ");
    for (i=0; i<p->n-p->m; i++)
      if (p->flags[i])
	printf("%11.2f", ucounts[i]);
    printf("\nu0 ");
    for (i=0; i<p->n-p->m; i++)
      if (p->flags[i])
	printf("%11.2f", u0counts[i]);
    printf("\nd0 ");
    for (i=0; i<p->n-p->m; i++)
      if (p->flags[i])
	printf("%11.2f", d0counts[i]);
    printf("\n");
  }

  i=0;
  while (!p->flags[i] && i<p->n-p->m) {
    i++;
  }

  if (i == p->n-p->m)
    retval = IMPOSSIBLE; /* should never happen */
  else if (ucounts[i] < SMALL)
    retval = IMPOSSIBLE; /* ideally shouldn't happen */
  else if (ncounts[i] < SMALL)
    retval = IMPOSSIBLE; /* shortcut */
  else {
    e = (double)ncounts[i] / ucounts[i];

    /* advance to next backoff level */
    oldi = i;
    do {
      i++;
    } while (!p->flags[i] && i<p->n-p->m);

    while (i<p->n-p->m && ucounts[i] > SMALL) {

      /* find next backoff level */
      j = i;
      do {
	j++;
      } while (!p->flags[j] && j<p->n-p->m);

      if (j < p->n-p->m)
	excess = u0counts[i] - u0counts[j];
      else
	excess = u0counts[i];

      if (d0counts[i] < 1.0) d0counts[i] = 1.0;
      
      lambda = excess / (u0counts[i] + C * d0counts[i]);
/*      lambda = u0counts[i] / (u0counts[i] + C * d0counts[i]);*/
      
/* for w|t... this seems to give unreasonably low lambdas */

      e = e * (1.0-lambda) + ncounts[i] / ucounts[i] * lambda;
      
      /* advance to next backoff level */
      oldi = i;
      i = j;
    }

    i = oldi;

/*    if (e > (ncounts[i] + 1) / ucounts[i]) {
      e = (ncounts[i] + 1) / ucounts[i];
      if (debug)
	printf ("clip at %d to %e (%e)\n", i, e, log10(e));
    }
    if (e < (ncounts[i] - 1) / ucounts[i]) {
      e = (ncounts[i] - 1) / ucounts[i];
      if (debug)
	printf ("clip at %d to %e (%e)\n", i, e, log10(e));
    }*/

    retval = log10(e);
  }

  if (debug) {
    printf(" => %f\n", retval);
  }
      
  /* Insert new value into cache */
  if (p->cache_ks) {
    if (!p->cache_ks[hash]) {
      p->cache_ks[hash] = malloc(sizeof(int) * p->n);
      if (p->cache_ks[hash] == NULL)
	return -1.0;
    }
    memcpy(p->cache_ks[hash], ks, sizeof(int) * p->n);
    p->cache_p[hash] = retval;
  }

  return retval;

}

double prob_cache_load(prob p) {
  int i, n = 0;
  for (i=0; i<p->cache_size; i++)
    if (p->cache_ks[i])
      n++;

  return (double)n/i;
}

void prob_stats(prob p) {
  fprintf (stderr, "Hit rate: %d / %d\n", p->hits, p->lookups);
}

int prob_set_mapping(prob p, char *names[], int n) {
  int i, j;

  for (i=0; i<p->n; i++) {
    for (j=0; j<n; j++) {
      if (!strcmp(p->names[i], names[j])) {
	p->mapping[i] = j;
	break;
      }
    }
  }
  return 0;
}

double prob_mapped_lookup(prob p, int *ks) {
  static int mks[MAX_DEPTH];
  int i;

  for (i=0; i<p->n; i++) {
    mks[i] = ks[p->mapping[i]];
  }
  return prob_lookup(p, mks);
}

int prob_mapped_count(prob p, int *ks, double delta) {
  static int mks[MAX_DEPTH];
  int i;

  for (i=0; i<p->n; i++) {
    mks[i] = ks[p->mapping[i]];
  }
  return prob_count(p, mks, delta);
}

