/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_GETOPT_H
# include <getopt.h>
#else
# include <unistd.h>
#endif

#include "prob.h"
#include "count.h"
#include "dict.h"

#define LINE_MAX 1024
#define MAX_TABLES 1024

dict d;
extern int debug;

#define COUNT_THRESHOLD 1e-15
float fuzz = 0.0;

prob p[MAX_TABLES];
int tables;
extern int rewrite_lambdas;

#define isspace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n')

void process_direct (char *buf) {
  int len = strlen(buf);
  int table;
  int i, j;
  prob cur;

  char *tokens[MAX_DEPTH];
  int ks[MAX_DEPTH];

  /* Determine table number */
  if (sscanf(buf, "%d", &table) == 1) {
    /* Now mark off all the tokens */
    i = 0;
    
    for (j=-1; j<p[table]->n; j++) { /* -1 to skip the table # */
      /* Skip initial whitespace */
      while (i < len && isspace(buf[i])) i++;
      if (i >= len) {
	fprintf(stderr, "event too short, table %d\n", table);
	exit(1);
      }
      if (j>=0) tokens[j] = buf+i;
      while (i < len && !isspace(buf[i])) i++;
      buf[i] = '\0';
      if (i < len) i++;
    }
    
    while (i < len && isspace(buf[i])) i++;
    if (i < len) {
      fprintf(stderr, "event too long, table %d\n", table);
      exit(1);
    }
    
    /* Lookup tokens in the dictionary */
    for (j=0; j<p[table]->n; j++) {
      ks[j] = dict_lookup(d, tokens[j]);
      if (ks[j] < 0) {
	ks[j] = dict_insert(d, tokens[j]);
	if (ks[j] < 0) {
	  fprintf (stderr, "Dictionary full\n");
	  exit(1);
	}
      }
    }
    if (prob_count(p[table], ks, 1.0) < 0) {
      fprintf(stderr, "prob_count() failed\n");
      exit(1);
    }
  }
}

void process_named(char *buf, double delta) {
  int ks[MAX_DEPTH];
  char *names[MAX_DEPTH], *values[MAX_DEPTH];
  int i, j, state;
  int k, n;
  prob cur;

  i = 0;
  k = 0;
  state = 0;
  while (buf[i] != '\0') {

    switch (state) {
    case 0: 
      if (!isspace(buf[i])) {
	j = i;
	state = 1;
      }
      break;

    case 1: 
      if (buf[i] == '=') {
	buf[i] = '\0';
	names[k] = buf+j;
	j = i+1;
	state = 2;
      }
      break;

    case 2: 
      if (isspace(buf[i])) {
	buf[i] = '\0';
	values[k] = buf+j;
	k++;
	state = 0;
      }
      break;
    }

    i++;

  }

  n = k;

  for (i=0; i<tables; i++) {
    for (cur = p[i]; cur != NULL; cur = cur->next) {
      for (j=0; j<cur->n; j++) {
	ks[j] = -1;
	for (k=0; k<n; k++)
	  if (!strcmp(names[k], cur->names[j])) {
	    ks[j] = dict_insert(d, values[k]);
	    break;
	  }
	if (ks[j] < 0)
	  fprintf(stderr, "missing value for %s\n", cur->names[j]);
      }
      if (prob_count(cur, ks, delta) < 0) {
	fprintf(stderr, "prob_count() failed\n");
	exit(1);
      }
    }
  }

}

int main (int argc, char *argv[]) {
  FILE *dictfp, *eventfp;
  char name[FILENAME_MAX];
  char buf[LINE_MAX];
  int i, table;
  prob cur;
  int events, dropped;
  double delta;

  int c;

  char *prefix, *oldprefix;
  int dict_size;

  srand(time(0));

  /* Process command line */

  while ((c = getopt(argc, argv, "r:f:")) != EOF)
    switch (c) {
    case 'r': rewrite_lambdas = 0; 
      oldprefix = strdup(optarg); break;
      break;
    case 'f': fuzz = atoi(optarg); break;
    }
  argc -= optind;
  argv += optind;

  if (argc != 2 || (dict_size = atoi(argv[1])) <= 0) {
    fprintf(stderr, "usage: trmodel <profile> <dictsize>\n");
    exit(1);
  }
  prefix = argv[0];

  /* If first arg ends in ".pro" chop it off */
  i = strlen(prefix);
  if (i > 4 && !strcmp(prefix + i - 4, ".pro"))
    prefix[i-4] = '\0';
  if (!rewrite_lambdas) {
    i = strlen(oldprefix);
    if (i > 4 && !strcmp(oldprefix + i - 4, ".pro"))
      oldprefix[i-4] = '\0';
  }

  tables = read_profile(prefix, p, MAX_TABLES);
  if (!rewrite_lambdas) {
    table = 0;
    for (i=0; i<tables; i++)
      for (cur=p[i]; cur != NULL; cur = cur->next) {
	sprintf(name, "%s.%d", oldprefix, table);
	prob_read_lambdas(cur, name);
	table++;
      }
  }

/*  eventfp = fopen("events", "r");*/
  eventfp = stdin;

  if (rewrite_lambdas)
    d = new_dict(dict_size);
  else {
    sprintf(name, "%s.dict", oldprefix);
    dictfp = fopen (name, "r");
    d = read_dict(dictfp);
    fclose(dictfp);
  }

  events = dropped = 0;
  /* OK, start counting */
  while (1) {
    fgets(buf, sizeof(buf), eventfp);

    if (feof(eventfp))
      break;

    events++;
    if (events % 100000 == 0) {
      if (dropped)
	fprintf(stderr, "%d events processed (%d dropped).\n", events, dropped);
      else
	fprintf(stderr, "%d events processed.\n", events);
    }

    if (sscanf(buf, "%lg%n", &delta, &i) < 1) {
      fprintf(stderr, "warning: missing delta\n");
      delta = 1.0;
    }

    delta += fuzz*rand()/RAND_MAX;

/*    process_direct(buf);*/

    if (!rewrite_lambdas || delta >= COUNT_THRESHOLD /* && delta <= 1.0*/)
      process_named(buf+i, delta);
    else
      dropped++;
    
  }

  printf("\nCleaning up...\n"); fflush(stdout);
  
  /* Rewrite directory file */
  sprintf(name, "%s.dict", prefix);
  dictfp = fopen(name, "w");
  dict_write(d, dictfp);
  fclose(dictfp);

  table = 0;
  for (i=0; i<tables; i++)
    for (cur=p[i]; cur!=NULL; cur=cur->next) {
      sprintf(name, "%s.%d", prefix, table);
      prob_write(cur, name);
      table++;
    }

  printf("Done.\n");

  exit(0);
}

