/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

/* add code and root to node_hash */

#include <config.h>

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include <unistd.h>
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#include "parser.h"
#include "model.h"
#include "prob.h"

#include "node.h"
#include "elem_label.h"
#include "elem_node.h"
#include "treebank_label.h"

#include "grammar.h"
#include "lexicon.h"


#define FOREACH(x,y) for((x)=(y); (x) != NULL; (x)=(x)->next)

#define LABEL_HASH_MAX 97

item chart[MAX_WORDS][MAX_WORDS][LABEL_HASH_MAX], oldchart[LABEL_HASH_MAX];
int spans[MAX_WORDS][MAX_WORDS];

float maxscore[MAX_WORDS][MAX_WORDS];
int total_items = 0, excluded = 0, updated = 0, pruned = 0;

int unknown;

/* Verbosity options */
int grammar_dump = 0, stats = 1, chart_dump = 0, chart_size_dump = 0;
int record_winners = 0;
int trace = 0;
int show_details = 0, show_all_parses = 0, show_hidden = 0;
int hwa = 0, charniak = 0;

/* Options for EM reestimation */
int compute_inside = 0;

/* For model.c */
extern char prefix[], derivation_filename[];
extern int derivation_dump, left_to_right, code_level;

/* For prob.c */
extern int debug;
extern float C;

/* This flag controls whether to use Mike Collins' punctuation rule. */
char *commastr = NULL;

lexicon words;

int max_cell_size = -1, tree_threshold = 2, word_threshold = 6, replace_tags = 1;
float beam_width = -4.0;
int skip_length = 100, punt_length = -1;

/* The parser. Described in (Schabes and Waters, 1993), adapted for the
   notion of TAG/TIG derivation of (Schabes and Shieber, 1994). */

/* Bugs: 

   Doesn't work for all grammars, namely, those that have substitution
   nodes without siblings (see below).

   Doesn't handle empty terminal nodes.

*/

item add(node n, code_t code, int i, int k, float dp, int action, item ant1, item ant2);
void close_item(item curItem, int i, int k);

void reconstruct(item curItem, struct word *w[]);
void derivation(item curItem, node above);

void item_delete(item oldItem);
void item_listDelete(item oldItem);
void print_item(item curItem);

code_t compute_code(code_t code, node n);
void code_string(code_t code, char *s, int n);
int code_equal(code_t code1, code_t code2);

/* derivation printing functions */
void print_attachment(node n, node s);
void print_eod(void);

/* BUG: doesn't work right for spans with feet */

int compute_spans(node cur, int start, int n) {
  int i, j, stop;

  stop = start;
  if (node_n_children(cur) == 0) {
    stop = start+1;
  } else {
    for (i=0; i<node_n_children(cur); i++)
      stop = compute_spans(node_child(cur, i), stop, n);
  }

  /* Clear all crossing spans */
  for (i=0; i<start; i++)
    for (j=start+1; j<stop; j++)
      spans[i][j] = 0;
  for (i=start+1; i<stop; i++)
    for (j=stop+1; j<n+1; j++)
      spans[i][j] = 0;

  return stop;
}

float beam(int span) {
  return beam_width;
}

void punt(struct word *w[], int n) {
  int i;
  static char tagbuf[100];

  if (n>0) {
    printf ("(S ");
    for (i=0; i<n; i++) {
      dict_rlookup(d, w[i]->t[0], tagbuf, sizeof(tagbuf));
      if (hwa && i<n-1) /* strict right-branching */
	printf (" (NP (%s %s) ", tagbuf, w[i]->word);
      else
	printf (" (%s %s)", tagbuf, w[i]->word);
    }
    if (hwa)
      for (i=0; i<n-1; i++)
	printf (")");
    printf (")\n");
    fflush(stdout);

    if (derivation_dump) {
      for (i=0; i<n; i++)
	print_attachment(NULL, NULL);
      print_eod();
    }
  }
}

int partition (float *a, int n, int pivot) {
  int i, j;
  float temp;
  float x = a[pivot];

  i = 0;
  j = n-1;

  while (a[i] >= x && i < n)
    i++;

  while (a[j] <= x && j >= 0)
    j--;

  while (i < j) {

    temp = a[i];
    a[i] = a[j];
    a[j] = temp;

    while (a[i] >= x && i < n)
      i++;

    while (a[j] <= x && j >= 0)
      j--;

  }

  /* pivot didn't move -- place it at boundary */
  if (pivot < j) {
    temp = a[pivot];
    a[pivot] = a[j];
    a[j] = temp;

    pivot = j;
  } else if (pivot > i) {
    temp = a[pivot];
    a[pivot] = a[i];
    a[i] = temp;

    pivot = i;
  }

  return pivot;
}

int float_compare(float *px, float *py) {
  if (*px < *py)
    return 1;
  if (*px > *py)
    return -1;
  return 0;
}

float order (float *a, int n, int k) {
/*
   int pos;
  if (n == 1)
    return a[0];

  if (k >= n)
    k = n-1;

  pos = partition(a, n, 0);

  if (k < pos)
    return order (a, pos, k);
  else if (k > pos)
    return order (a+pos+1, n-pos-1, k-pos-1);
  else
    return a[pos];*/

  if (k >= n)
    k = n-1;

  qsort(a, n, sizeof(float), (int (*)(const void *, const void *))float_compare);

  return a[k];
}

/* This function must group nodes of the same label together. */
int node_hash(node n) {
  return node_label(n) % LABEL_HASH_MAX;
}

int wild_hash(int dir) {
  static int left_wild, right_wild, initialized = 0;

  if (!initialized) {
    left_wild = dict_lookup(d, "<");
    right_wild = dict_lookup(d, ">");
  }

  if (dir == LEFT)
    return left_wild % LABEL_HASH_MAX;
  else if (dir == RIGHT)
    return right_wild % LABEL_HASH_MAX;
  else
    return -1;
}

int node_match(node n, int w, int tags[], int n_t) {
  int tag = node_label(node_parent(n));
  int i;

  if (w != node_label(n))
    return 0;
  for (i=0; i<n_t; i++)
    if (tag == tags[i])
      return 1;
  return 0;
}

int seed(node root, struct word *w[], int n) {
  int i;
  if (node_n_children(root) == 0) {
    if (node_type(root) == FOOT) {
      for (i=0; i<=n; i++)
	add(root, CODE_STOP, i, i, 0.0, ACTION_POP, NULL, NULL);
    } else if (node_type(root) == FRONTIER && node_empty(root)) {
      for (i=0; i<=n; i++)
      add(root, CODE_STOP, i, i, 0.0, ACTION_NODE, NULL, NULL);
    } else if (node_type(root) == FRONTIER) {
      for (i=0; i<n; i++)
	if (node_label(root) == w[i]->ow &&
	    node_match(root, w[i]->ow, w[i]->t, w[i]->n_t))
	  add(root, CODE_STOP, i, i+1, 0.0, ACTION_NODE, NULL, NULL);
    }
  } else {
    for (i=0; i<node_n_children(root); i++)
      seed(node_child(root, i), w, n);
  }

  return 0;
}

int word_comma(struct word *w) {
  int i, j, n;
  static int initialized = 0;
  static int *commas, n_commas;

  if (!initialized) {
    /* Parse comma string */
    
    if (commastr == NULL)
      commastr = strdup(",|.|:|;|-|--|...");
    
    n = strlen(commastr);
    if (n == 0) {
      n_commas = 0;
    } else {
      n_commas = 1;
      for (i=0; i<n; i++) {
	if (commastr[i] == '|') {
	  commastr[i] = '\0';
	  n_commas++;
	}
      }
      commas = malloc(n_commas * sizeof(int));
      commas[0] = dict_lookup(d, commastr);
      j = 1;
      for (i=0; i<n; i++) {
	if (commastr[i] == '\0') {
	  commas[j] = dict_lookup(d, commastr+i+1);
	  j++;
	}
      }
    }
    initialized = 1;
  }
  
  for (i=0; i<n_commas; i++)
    if (w->w == commas[i])
      return 1;
  return 0;
}

float parse(struct word *w[], int n, list_node grammar) {
  int i, j, k, l, nt;
  node parent;
  item curItem, jItem, best;
  item *pCurItem;
  int root;
  float p, pi, bestpi;
  list_node li;

  float threshold;

#define MAX_SCORES 10000
  static float scores[MAX_SCORES];
  int si;

  if (stats) {
    total_items = excluded = updated = pruned = 0;
  }

  for (i=0; i<=n; i++)
    for (k=0; k<=n; k++) {
      maxscore[i][k] = IMPOSSIBLE;
      for (nt=0; nt<LABEL_HASH_MAX; nt++)
	chart[i][k][nt] = NULL;
    }

  /* Axioms */

  if (trace)
    printf("\nAxioms\n===========\n");

  for (li = grammar; li != NULL; li = li->next)
    seed(tree_root(li->contents), w, n);

  if (trace)
    printf("Beginning parsing.\n");

  for (l=1; l<=n; l++)
    for (i=0; i<=n-l; i++) {
      k = i+l;

      if (trace)
	printf("Computing <%d, %d>\n", i, k);

      if (!spans[i][k])
	continue;

      for (j=i+1; j<=k-1; j++) {

	/* Punctuation rule */
/*	if (word_comma(w[j]) || (word_comma(w[j-1]) && j != i+1))
	  if (!(k == n || word_comma(w[k]) || word_comma(w[k-1])))
	    continue;*/
	if (word_comma(w[j-1]) /*&& j != i+1*/ && k != n && !word_comma(w[k]))
	    continue;
	
	if (trace)
	  printf("j = %d\n", j);
	for (nt=0; nt<LABEL_HASH_MAX; nt++)
	FOREACH (curItem, chart[i][j][nt]) {
/*
printf("Examining ");
  putchar('[');
  node_print(curItem->n);
  printf(", %d, %d, %d, %f]\n", curItem->code, i, j, curItem->p);
*/
	    
	  root = node_isroot(curItem->n);
	  parent = node_parent(curItem->n);

	  /* Left modifier adjunction */
	  /* If the node is the root of a left modifier tree,
	     scan chart[j][k][label] for adjunction nodes */
	  if (code_equal(curItem->code, CODE_STOP) &&
	      root &&
	      tree_type(curItem->n) == LEFTMOD) {
	    FOREACH (jItem, chart[j][k][wild_hash(LEFT)])
	      if (!code_equal(jItem->code, CODE_STOP) &&
		  node_adjoinable(curItem->n, jItem->n)) {
		add(jItem->n, compute_code(jItem->code, curItem->n), i, k, 
		    p_adj(curItem->n, jItem->n, jItem->code),
		    ACTION_PUSH, jItem, curItem);
	      }

	    /* Left predicative adjunction */
	    /* This block is nearly the same as the previous one
	       but they are separate to keep things sane, and to
	       faciliate tinkering with the notion of derivation. 
	       The notion of derivation here is that all modifier
	       adjunction precedes predicative adjunction. */
	  } else if (code_equal(curItem->code, CODE_STOP) &&
	      root &&
	      tree_type(curItem->n) == LEFTPRED) {
	    FOREACH (jItem, chart[j][k][node_hash(curItem->n)])
	      if (!code_equal(jItem->code, CODE_STOP) &&
		  node_adjoinable(curItem->n, jItem->n))
		add(jItem->n, CODE_STOP, i, k, 
		    p_adj(curItem->n, jItem->n, jItem->code),
		    ACTION_PUSH, jItem, curItem);

	    /* Substitution */
	    /* If the node has a right sister,
	       scan chart[j][k][label] for it */
	  } else if (code_equal(curItem->code, CODE_STOP) &&
		     !root &&
		     node_right(curItem->n) != NULL) {
	    if (node_type(node_right(curItem->n)) == SUBST) {
	      FOREACH (jItem, chart[j][k][node_hash(node_right(curItem->n))]) {
		if (code_equal(jItem->code, CODE_STOP) &&
		    node_isroot(jItem->n) &&
		    tree_type(jItem->n) == INITIAL &&
		    node_substitutable(jItem->n, node_right(curItem->n)))
		  add(parent, CODE_START, i, k,
		      p_subst(jItem->n, node_right(curItem->n)),
		      ACTION_NODE, curItem, jItem);
	      }

	      /* Concatenation */
	    } else {
	      FOREACH (jItem, chart[j][k][node_hash(node_right(curItem->n))]) {
		if (code_equal(jItem->code, CODE_STOP) &&
		    jItem->n == node_right(curItem->n)) {
		  add(node_parent(jItem->n), CODE_START, i, k, 
		      0.0, ACTION_NODE, curItem, jItem);
		}
	      }
	    }
	  }

	}

	for (nt=0; nt<LABEL_HASH_MAX; nt++)
	FOREACH (curItem, chart[j][k][nt]) {

	  root = node_isroot(curItem->n);
	  parent = node_parent(curItem->n);

	  /* Right modifier adjunction */
	  if (code_equal(curItem->code, CODE_STOP) &&
	      root &&
	      tree_type(curItem->n) == RIGHTMOD) {
	    FOREACH (jItem, chart[i][j][wild_hash(RIGHT)])
	      if (!code_equal(jItem->code, CODE_STOP) &&
		  node_adjoinable(curItem->n, jItem->n)) {
		add(jItem->n, compute_code(jItem->code, curItem->n), i, k,
		    p_adj(curItem->n, jItem->n, jItem->code),
		    ACTION_PUSH, jItem, curItem);
	      }

	    /* Right predicative adjunction */
	  } else if (code_equal(curItem->code, CODE_STOP) &&
	      root &&
	      tree_type(curItem->n) == RIGHTPRED) {
	    FOREACH (jItem, chart[i][j][node_hash(curItem->n)])
	      if (!code_equal(jItem->code, CODE_STOP) &&
		  node_adjoinable(curItem->n, jItem->n)) {
		add(jItem->n, CODE_STOP, i, k,
		    p_adj(curItem->n, jItem->n, jItem->code),
		    ACTION_PUSH, jItem, curItem);
	      }

	    /* Substitution */
	  } else if (code_equal(curItem->code, CODE_STOP) &&
		     !root &&
		     node_left(curItem->n) != NULL &&
		     node_type(node_left(curItem->n)) == SUBST) {
	    FOREACH(jItem, chart[i][j][node_hash(node_left(curItem->n))])
	      if (code_equal(jItem->code, CODE_STOP) &&
		  node_isroot(jItem->n) &&
		  tree_type(jItem->n) == INITIAL &&
		  node_substitutable(jItem->n, node_left(curItem->n)))
		add(parent, CODE_START, i, k,
		    p_subst(jItem->n, node_left(curItem->n)),
		    ACTION_NODE, jItem, curItem);
	  }

	}

      }

      /* Second pruning pass */

      if (max_cell_size > 0) {
	si = 0;
	for (nt=0; nt<LABEL_HASH_MAX; nt++) {
	  FOREACH (curItem, chart[i][k][nt])  {
	    if (si < MAX_SCORES)
	      scores[si] = curItem->p + curItem->p_prior;
	    si++;
	  }
	}

	if (si >= MAX_SCORES) {
	  fprintf(stderr, "Warning: %d scores spanning <%d, %d>: threshold will be inaccurate\n", si, i, k);
	  si = MAX_SCORES;
	}
	
	threshold = order(scores, si, max_cell_size);
	
	if (threshold < maxscore[i][k] + beam(k-i))
	  threshold = maxscore[i][k] + beam(k-i);
      } else {
	
	threshold = maxscore[i][k] + beam(k-i);
      }
      
      if (trace) {
	printf ("Second pruning pass (threshold = %f)\n", threshold);
      }

      for (nt=0; nt<LABEL_HASH_MAX; nt++) {

	pCurItem = &chart[i][k][nt];
	curItem = *pCurItem;

	while (curItem != NULL) {
	  if (curItem->p + curItem->p_prior < threshold) {
	    *pCurItem = curItem->next;
	    if (trace) {
	      print_item(curItem);
	      printf(" pruned\n");
	    }
	    pruned++;
	    item_delete(curItem);
	  } else {
	    pCurItem = &curItem->next;
	  }
	  curItem = *pCurItem;
	}
      }

      /* Close up the newly generated items. */

      for (nt=0; nt<LABEL_HASH_MAX; nt++)
	oldchart[nt] = chart[i][k][nt];

      for (nt=0; nt<LABEL_HASH_MAX; nt++)
	FOREACH (curItem, oldchart[nt]) {
	  close_item(curItem, i, k);
	}

    }
  
  /* Scan chart[0][n] for initial roots */
  
  best = NULL; bestpi = IMPOSSIBLE;
  for (nt=0; nt<LABEL_HASH_MAX; nt++)
  FOREACH (curItem, chart[0][n][nt]) {
    if (code_equal(curItem->code, CODE_STOP) &&
	node_isroot(curItem->n) &&
	tree_type(curItem->n) == INITIAL) {
      pi = p_initial(curItem->n);
      if (show_all_parses) {
	printf ("%f + %f ", curItem->p, pi);
	reconstruct(curItem, w);
      }
      if (best == NULL || curItem->p + pi > best->p + bestpi) { 
	best = curItem;
	bestpi = pi;
      }
    }
  }

  if (best) {
    if (derivation_dump) {
      derivation(best, NULL);
      print_eod();
    }

    reconstruct(best, w);
    p = best->p + bestpi;
  } else {
    punt(w, n);
    p = IMPOSSIBLE;
  }

  if (stats) {
    int sum = 0, max = 0, support = 0, imax, kmax, ntmax;

    for (i=0; i<=n; i++)
      for (k=0; k<=n; k++) 
	for (nt=0; nt<LABEL_HASH_MAX; nt++) {
	j = 0;
	FOREACH(curItem, chart[i][k][nt])
	  j++;
	if (j>max) {
	  max = j; imax = i; kmax = k; ntmax = nt;
	}
	if (j>0)
	  support++;
	sum += j;
      }
    fprintf(stderr, "Sentence length: %d Average cell size: %d maximum: %d total: %d\n", n, support ? sum/support : 0, max, sum);
    fprintf(stderr, "Total items generated: %d excluded: %d updated: %d pruned: %d efficiency: %f\n", total_items, excluded, updated, pruned, total_items ? (float)sum/total_items : 0.0);
    node_stats();

  }

  if (chart_dump) {
    for (i=0; i<=n; i++)
      for (k=0; k<=n; k++) {

	printf ("\n<%d,%d>\n=======\n", i, k);

	for (nt=0; nt<LABEL_HASH_MAX; nt++)
	  FOREACH(curItem, chart[i][k][nt]) {
	    print_item(curItem);
	    printf("\n");
	  }
      }
  }

  if (chart_size_dump) {

    printf("Cell sizes:\n");
    for (i=0; i<=n; i++) {
      for (k=0; k<=n; k++) {
	j = 0;
	for (nt=0; nt<LABEL_HASH_MAX; nt++)
	  FOREACH(curItem, chart[i][k][nt])
	    j++;
	if (j>0)
	  printf("%8d", j);
	else
	  printf("        ");
      }
      printf("\n");
    }
  }


  /* Clean up */

  for (i=0; i<=n; i++)
    for (k=0; k<=n; k++) 
      for (nt=0; nt<LABEL_HASH_MAX; nt++) {
      item_listDelete(chart[i][k][nt]);
/*      chart[i][k][nt] = NULL;*/ /* this gets done next time around */
    }

  return p;

}

void close_item(item curItem, int i, int k) {
  item jItem;

  if (curItem == NULL)
    return;

  if (code_equal(curItem->code, CODE_STOP)) {
      
    if (node_isroot(curItem->n)) {
      if (tree_type(curItem->n) == INITIAL) {
	/* Substitution */
/*	li = new_leafIterator(SUBST, curItem->n);
	curNode = leafIterator_next(li);
	while (curNode != NULL) {
	  add(curNode, CODE_START, i, k, p_subst(curItem->n, curNode), ACTION_NONE, curItem, NULL);
	  curNode = leafIterator_next(li);
	}*/
      }
      
    } else if (node_left(curItem->n) == NULL) {
      if (node_right(curItem->n) == NULL) {

	/* Unary concatenation */
	close_item(add (node_parent(curItem->n), CODE_START, i, k, 0.0, ACTION_NODE, curItem, NULL), i, k);
      } else if (node_type(node_right(curItem->n)) == FOOT || node_empty(node_right(curItem->n))) {

	/* Foot concatenation */
	FOREACH (jItem, chart[k][k][node_hash(node_right(curItem->n))])
	  if (code_equal(jItem->code, CODE_STOP) && jItem->n == node_right(curItem->n)) {
	    close_item(add (node_parent(curItem->n), CODE_START, i, k, 0.0, ACTION_NODE, curItem, jItem), i, k);
	    break;
	  }
      }
    } else if (node_type(node_left(curItem->n)) == FOOT || node_empty(node_left(curItem->n))) {

      /* Foot concatenation */
      FOREACH (jItem, chart[i][i][node_hash(node_left(curItem->n))])
	if (code_equal(jItem->code, CODE_STOP) && jItem->n == node_left(curItem->n)) {
	  close_item(add (node_parent(curItem->n), CODE_START, i, k, 0.0, ACTION_NODE, jItem, curItem), i, k);
	  break;
	}
    }
  } else {
    /* No adjunction */
    close_item(add (curItem->n, CODE_STOP, i, k, p_adj(NULL, curItem->n, curItem->code), ACTION_NONE, curItem, NULL), i, k);
  }

}

item new_item(node n, code_t code, float p, int action, item ant1, item ant2) {
  item retval;

  retval = malloc(sizeof(*retval));
  retval->n = n;
  retval->code = code;
  retval->p = p;
  retval->next = NULL;
  retval->action = action;
  retval->ant1 = ant1;
  retval->ant2 = ant2;

  retval->p_inside = retval->p_outside = 0.0;

  return retval;
}

void item_delete(item oldItem) {
  free (oldItem);
}

void item_listDelete(item oldItem) {
  item cur, next;

  cur = oldItem;

  while (cur != NULL) {
    next = cur->next;
    item_delete(cur);
    cur = next;
  }
}

item copy_item(item oldItem) {
  item retval;

  retval = malloc(sizeof(*retval));
  retval->n = oldItem->n;
  retval->code = oldItem->code;
  retval->p = oldItem->p;
  retval->p_prior = oldItem->p_prior;
  retval->i = oldItem->i;
  retval->j = oldItem->j;

  retval->next = oldItem->next;
  retval->action = oldItem->action;
  retval->ant1 = oldItem->ant1;
  retval->ant2 = oldItem->ant2;

  return retval;
}

void print_item(item curItem) {
  char buf[80];
  printf("%x:", (unsigned int)curItem);
  printf("<%d, %d> [", curItem->i, curItem->j);
  node_print(curItem->n);
  code_string(curItem->code, buf, sizeof(buf));
  printf(", %s, %f/%f; %x %x]", buf, curItem->p, curItem->p + curItem->p_prior, curItem->ant1, curItem->ant2);
}

item add(node n, code_t code, int i, int k, float dp, int action, item ant1, item ant2) {
  item curItem;
  int found = 0;
  float score, p, pp;
  char buf[80];

  total_items++;

  p = dp + (ant1 ? ant1->p : 0.0) + (ant2 ? ant2->p : 0.0);
  pp = p_prior(n, code);

  score = p + pp;

  if (trace) {
    printf("[");
    node_print(n);
    code_string(code, buf, sizeof(buf));
    printf(", %s, %d, %d, %f/%f]", buf, i, k, p, score);
    if (ant1) {
      printf(" < ");
      print_item(ant1);
      if (ant2) {
	printf(", ");
	print_item(ant2);
      }
    }
  }

  if (score < beam(k-i) + maxscore[i][k]) {
    if (trace)
      printf (" pruned\n");
    excluded++;
    return NULL;
  }
   

  if (score > maxscore[i][k])
    maxscore[i][k] = score;

  /* Scan chart[i][k] for (n, code) */
  FOREACH (curItem, chart[i][k][node_hash(n)]) {
    if (code_equal(curItem->code, code) && curItem->n == n) {
      found = 1;
      break;
    }
  }

  /* If present, set p to max and change antecedents */
  if (found) {
    updated++;

    if (compute_inside) {
      curItem->p_inside += pow(10.0, p);
    }
    
    if (p > curItem->p) {

      curItem->p = p;
      curItem->p_prior = pp;
      
      /* update antecedents */
      curItem->action = action;
      curItem->ant1 = ant1;
      curItem->ant2 = ant2;

    }

  } else if (!found) {

    /* Otherwise, add to bin */
    curItem = new_item(n, code, p, action, ant1, ant2);
    curItem->i = i;
    curItem->j = k;

    curItem->p_prior = pp;

    curItem->next = chart[i][k][node_hash(n)];
    chart[i][k][node_hash(n)] = curItem;

    if (trace)
      printf (" (%x)", (unsigned int)curItem);

  } 

  if (trace)
    printf ("\n");

  return curItem;

}

void record_winner(item curItem, int i, int j) {
    fprintf (stderr, "item: %d %f\n", j-i, maxscore[i][j] - (curItem->p + curItem->p_prior));
}

void reconstruct_helper(item curItem, item stack, struct word *w[]) {
  item newItem;
  int visible;

  if (curItem != NULL) {

    if (record_winners)
      record_winner(curItem, curItem->i, curItem->j);

    if (curItem->action == ACTION_NODE) {
      if (node_type(curItem->n) == FRONTIER) {
	if (node_label(curItem->n) == unknown)
	  printf("%s", w[curItem->i]->word);
	else
	  node_print(curItem->n);
      } else {

/*
printf(" <NODE ");
node_print(curItem->ant1->n);
if (curItem->ant2) {
  printf(" ");
  node_print(curItem->ant2->n);
}
printf("> ");

if (node_parent(curItem->ant1->n) != curItem->n)
printf("ERROR");
*/

	visible = node_visible(curItem->n);
	
/*	if ((tree_type(curItem->n) == LEFTMOD || tree_type(curItem->n) == RIGHTMOD) && node_isroot(curItem->n))
	  visible = 0;*/

	if (show_details || show_hidden)
	  visible = 1;

	if (hwa) {
	  visible = curItem->ant2 != NULL || node_type(curItem->ant1->n) == FRONTIER;
	}

	if (visible) {
	  printf("(");
	  node_print(curItem->n);
	  printf(" ");
	}
	reconstruct_helper (curItem->ant1, node_spine(curItem->ant1->n) ? stack : NULL, w);
	if (curItem->ant2 != NULL) {
	  printf(" ");
	  reconstruct_helper (curItem->ant2, node_spine(curItem->ant2->n) ? stack : NULL, w);
	}

	if (visible)
	  printf(")");
      }
    } else if (curItem->action == ACTION_PUSH) {

/*
printf(" <PUSH ");
node_print(curItem->ant1->n);
printf(" ");
node_print(curItem->ant2->n);
printf("> ");
*/

      if (tree_type(curItem->ant2->n) == LEFTPRED || tree_type(curItem->ant2->n) == RIGHTPRED) {

	/* Duplicate curItem and push */
	newItem = copy_item(curItem->ant1);
	newItem->next = stack;
	stack = newItem;
	
      }


      /* We treat modifier trees specially because there
	 is no pop */
      if (node_virtual(curItem->ant1->n) == RIGHT) {
	if (hwa)
	  printf("(> ");
	reconstruct_helper(curItem->ant1, stack, w);
      } else if (node_virtual(curItem->ant1->n) == LEFT && hwa) {
	printf("(< ");
      }

      reconstruct_helper(curItem->ant2, stack, w);

      if (node_virtual(curItem->ant1->n) == LEFT) {
	reconstruct_helper(curItem->ant1, stack, w);
	if (hwa)
	  printf(")");
      } else if (node_virtual(curItem->ant1->n) == RIGHT && hwa) {
	printf(")");
      }

    } else if (curItem->action == ACTION_POP) {

/*
printf(" <POP> ");
*/
      /* Pop */
      newItem = stack;
      stack = stack->next;

      reconstruct_helper(newItem, stack, w);
    } else if (curItem->action == ACTION_NONE) {
/*
printf(" <NONE ");
node_print(curItem->ant1->n);
printf("> ");
*/
      reconstruct_helper(curItem->ant1, stack, w);
    }
  }
}

void reconstruct(item curItem, struct word *w[]) {
  reconstruct_helper(curItem, NULL, w);
  printf("\n");
  fflush(stdout);
}

void derivation(item curItem, node above) {

  if (curItem == NULL)
    return;

  if (derivation_dump >= 2) {
    print_item(curItem);
    printf(" < ");
    if (curItem->ant1)
      print_item(curItem->ant1);
    if (curItem->ant2)
      print_item(curItem->ant2);
    printf("\n");
  }

  if (curItem->action == ACTION_PUSH && 
      (tree_type(curItem->ant2->n) == LEFTMOD || tree_type(curItem->ant2->n) == LEFTPRED))
    derivation(curItem->ant2, curItem->n);

  if (curItem->ant1 && node_isroot(curItem->ant1->n) && code_equal(curItem->ant1->code, CODE_STOP))
    derivation(curItem->ant1, node_left(curItem->ant2->n)); /* subst */
  else
    derivation(curItem->ant1, above);

  if (curItem->action != ACTION_PUSH) {
    if (curItem->ant2 && node_isroot(curItem->ant2->n) && code_equal(curItem->ant2->code, CODE_STOP))
      derivation(curItem->ant2, node_right(curItem->ant1->n));
    else
      derivation(curItem->ant2, above);
  }

  /* leaf node */
  if (curItem->action == ACTION_NODE && curItem->ant1 == NULL) {
    print_attachment(curItem->n, above);
  }

  if (curItem->action == ACTION_PUSH && 
      (tree_type(curItem->ant2->n) == RIGHTMOD || tree_type(curItem->ant2->n) == RIGHTPRED))
    derivation(curItem->ant2, curItem->n);

  return;

}

int main (int argc, char *argv[]) {
  struct word **w, *w1;
  int n;
  int i, j, k;
  list_node grammar;
  time_t start;
  int c;
  char *options, *value;
  FILE *fp;
  static char filename[FILENAME_MAX];

  static char *verbopts[] = {
    "grammar",
    "chart",
    "trace",
    "prob",
    "problevels",
    NULL
  };
  
  start = time(NULL);

  /* Process command-line options */
  while ((c = getopt(argc, argv, "b:t:m:c:d:C:P:hp:s:v:w:nRH")) != EOF)
    switch (c) {
    case 'b': beam_width = -fabs(atof(optarg)); break;
    case 't': tree_threshold = atoi(optarg); break;
    case 'w': word_threshold = atoi(optarg); break;
    case 'n': replace_tags = 0; break;
    case 'm': skip_length = atoi(optarg); 
      if (skip_length >= MAX_WORDS) 
	skip_length=MAX_WORDS-1; 
      break;
    case 'M': max_cell_size = atoi(optarg); break;
    case 'c': commastr = strdup(optarg); break;
    case 'P': punt_length = atoi(optarg); break;

    case 'h': left_to_right = 0; break;
    case 's': code_level = atoi(optarg); break;
    case 'p': strncpy(prefix, optarg, FILENAME_MAX); break;
    case 'd': derivation_dump = 1; strncpy(derivation_filename, optarg, FILENAME_MAX); break;

    case 'C': C = atof(optarg); break;
    case 'H': hwa = 1; break;

#ifdef HAVE_GETSUBOPT
    case 'v': 
      options = optarg;
      show_details = 1;
      while (*options != '\0') {
	switch (getsubopt(&options, verbopts, &value)) {
	case 0: grammar_dump = 1; break;
	case 1: chart_dump = chart_size_dump = 1; break;
	case 2: trace = 1; break;
	case 3: debug = 1; break;
	case 4: debug = 2; break;
	}
      }
#endif

    }

  model_init();

  w = malloc(skip_length * sizeof(struct word *));

  unknown = dict_lookup(d, "*UNKNOWN*");

  /* Read in .lex file */
  words = new_lexicon((COMPARE)word_compare, (HASH)word_hash);
  sprintf(filename, "%s.lex", prefix);
  fp = fopen(filename, "r");
  lexicon_read(words, fp, (READ)read_word);
  fclose(fp);

  if (stats)
    fprintf(stderr, "(%d s) Done initializing\n", time(NULL)-start);

  k=0;

  while (!feof(stdin)) {

    n = read_sent(stdin, w, skip_length);
      
    if (stats)
      fprintf(stderr, "Read sentence:");
    
    for (i=0; i<n; i++) {
      w1 = lexicon_lookup(words, w[i]);
      if (w1 != NULL && replace_tags && !hwa)
	word_merge(w[i], w1);
      if (hwa || lexicon_count(words, w[i]) < word_threshold)
	w[i]->w = unknown;
      
      if (stats) {
	fprintf(stderr, " ");
	write_word(stderr, w[i]);
      }
      
    }
    
    if (stats)
      fprintf(stderr, "\n");
    
    if (n < 0) {
      if (stats)
	fprintf(stderr, "(%d s) Skipping sentence\n", time(NULL)-start);
    } else if (punt_length > 0 && n > punt_length) {
      if (stats)
	fprintf(stderr, "(%d s) Punting on sentence\n", time(NULL)-start);
      punt(w, n);
    } else if (n <= skip_length) {

      grammar = growTrees(w, n);

      for (i=0; i<n+1; i++)
	for (j=0; j<n+1; j++)
	  spans[i][j] = 1;
/*	compute_spans(derived_tree, 0, n);*/

      parse(w, n, grammar);

      if (stats)
	fprintf(stderr, "(%d s) Done parsing\n", time(NULL)-start);
      
      deleteTrees(grammar);
    }

    for (i=0; i<n; i++) {
      free(w[i]->word);
      free(w[i]->t);
      free(w[i]);
    }
      
    k++;

  }

  done();
}


