/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <string.h>

#include "parser.h"
#include "dict.h"

#include "treebank_label.h"

extern dict d;

#define MAX_TAGS 100

#define isspace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n')

int word_compare(struct word *w1, struct word *w2) {
  return strcmp(w1->word, w2->word);
}

unsigned int word_hash(struct word *w) {
  char *s = w->word;
  unsigned int hash = 0;
  while (*s != '\0') {
    hash = hash*2+*s;
    s++;
  }
  return hash;
}

int word_merge(struct word *w1, struct word *w2) {
  static int t[MAX_TAGS];
  int n_t, i, j, found;

  for (i=0; i<w1->n_t && i<MAX_TAGS; i++)
    t[i] = w1->t[i];
  n_t = w1->n_t;

  for (i=0; i<w2->n_t; i++) {
    found = 0;
    for (j=0; j<w1->n_t; j++)
      if (w1->t[j] == w2->t[i])
	found = 1;
    if (!found && n_t<MAX_TAGS) {
      t[n_t] = w2->t[i];
      n_t++;
    }
  }

  free(w1->t);
  w1->t = malloc(n_t * sizeof(int));
  for (i=0; i<n_t; i++)
    w1->t[i] = t[i];
  w1->n_t = n_t;

  return 0;
}

/* Input routines */

/* Input format:

   (WORD TAG TAG ...) (WORD TAG TAG ...) ...
   (WORD TAG TAG ...) ...
   ...

   */

int read_token(FILE *fp, char *buf, int n, const char *delim) {
  int i = 0, c;

  /* Skip over initial isspace. */
  c = getc(fp);
  while (c != EOF && isspace(c))
    c = getc(fp);

  while (c != EOF && !strchr(delim, c) && i<n) {
    buf[i] = c;
    i++;
    c = getc(fp);
  }
  if (ferror(fp)) perror("read_token");

  if (i<n) buf[i] = '\0';

  return c;
}

void write_word(FILE *fp, struct word *w) {
  static char buf[100];
  int i;

  dict_rlookup(d, w->t[0], buf, sizeof(buf));
  fprintf(fp, "(%s", buf);
  for (i=1; i<w->n_t; i++) {
    dict_rlookup(d, w->t[i], buf, sizeof(buf));
    fprintf(fp, "|%s", buf);
  }
  dict_rlookup(d, w->w, buf, sizeof(buf));
  fprintf(fp, " %s", buf);
  if (!!strcmp(buf, w->word)) {
    fprintf(fp, "[%s]", w->word);
  }
  fprintf(fp, ")");
}

struct word * read_word(FILE *fp) {
  static int tags[MAX_TAGS];
  int i, j, result, c;
  static char buf[MAX_LENGTH];
  struct word *w;

  w = malloc(sizeof (struct word));

  /* Skip over initial isspace. */
  c = getc(fp);
  while (c != EOF && isspace(c))
    c = getc(fp);
  if (ferror(fp)) perror("read_word");

  if (c != '(')
    return NULL;

  i = 0;

  result = read_token(fp, buf, sizeof(buf), "| ");
  while (result >= 0) {

    tags[i] = dict_lookup(d, buf);
    i++;
    if (result == ' ')
      break;
    result = read_token(fp, buf, sizeof(buf), "| ");
  }

  w->n_t = i;
  w->t = malloc(i*sizeof(int));
  for (j=0; j<i; j++)
    w->t[j] = tags[j];
  
  read_token(fp, buf, sizeof(buf), ")");

  /* remove any trailing isspace */
  for (i=0; buf[i] != '\0'; i++)
    if (isspace(buf[i])) {
      buf[i] = '\0';
      break;
    }

  w->word = strdup(buf);
  w->ow = w->w = dict_insert(d, buf);
  w->derived_node = NULL;

  return w;
}

int read_sent(FILE *fp, struct word *w[], int n) {
  int c, i, error = 0;

  i = 0;
  while (1) {
    c = getc(fp);
    switch (c) {
    case '\n': case EOF: 
      if (ferror(fp)) perror("read_sent");
      return error?-1:i;
    case ' ': continue;
    case '(':
      if (i<n) {
	ungetc('(', fp);
	w[i] = read_word(fp);
	i++;
      } else {
	error = 1;
      }
    }
  }

  return i;
}

int tree_to_sent_helper(node cur, struct word *w[], int n, int m) {
  int i;

  if (node_n_children(cur) == 0) {
    w[m] = malloc(sizeof (struct word));
    w[m]->ow = w[m]->w = dict_insert(d, ((treebank_label)cur->contents)->label);
    w[m]->word = strdup(((treebank_label)cur->contents)->label);
    w[m]->n_t = 1;
    w[m]->t = malloc(sizeof(int));
    w[m]->t[0] = dict_lookup(d, ((treebank_label)node_parent(cur)->contents)->label);
    w[m]->derived_node = cur;

    return m+1;
  } else {
    for (i=0; i<node_n_children(cur); i++)
      m = tree_to_sent_helper(node_child(cur, i), w, n, m);
    return m;
  }
}

int tree_to_sent(node root, struct word *w[], int n) {
  return tree_to_sent_helper(root, w, n, 0);
}

