#include <config.h>

#include <unistd.h>
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "chart.h"

#include "model.h"

#include "node.h"
#include "elem_label.h"
#include "elem_node.h"
#include "treebank_label.h"

#include "grammar.h"
#include "lexicon.h"

#include "hash_map.h"

#include "prob.h"

#define MAX_WORDS 250

int alloc = 0;

int show_details = 1;
int tree_threshold = 1, word_threshold = 6;
int grammar_dump = 0, stats = 1, trace = 0;
extern int left_to_right;
extern int debug; /* from prob.c */
extern float C;
int unknown, nil, stop;
int left_wild, right_wild;
lexicon words;

extern list_node subgrammar;
list_node subst_nodes;

#define exp10(x) (pow(10.0, (x)))

struct inference * new_inference(int op, item ant1, item ant2) {
  struct inference *pip;
  pip = malloc(sizeof(*pip));
  pip->op = op;
  pip->p = 1.0;
  pip->ant1 = ant1;
  pip->ant2 = ant2;

  return pip;
}

void inference_delete(struct inference *pip) {
  free(pip);
}

item new_item(node n, node dn1, node dn2, node df, code_t code, int op, item ant1, item ant2) {
  item it;
  it = malloc(sizeof(*it));

  it->n = n;
  it->dn1 = dn1;
  it->dn2 = dn2;
  it->df = df;
  it->code = code;

  it->flag = 0;
  it->p_inside = 0.0;
  it->p_outside = 0.0;

  it->inferences = new_list_node(new_inference(op, ant1, ant2), NULL);

  return it;
}

void item_delete(item it) {
  list_node li;
  if (it == NULL)
    return;

  for (li = it->inferences; li != NULL; li = li->next)
    inference_delete(li->contents);
  list_delete(it->inferences);
  free(it);
}

void item_print(FILE *fp, item it) {
  char buf1[100], buf2[100];

  if (it->n)
    elem_label_to_string(it->n->contents, buf1, sizeof(buf1));
  else
    strcpy(buf1, "NIL");
  code_string(it->code, buf2, sizeof(buf2));
  fprintf(fp, "[%s, %s, %s, %s, %s] (a_hash %d)", 
	  buf1,
	  buf2,
	  it->dn1 ? ((treebank_label)it->dn1->contents)->label : "NIL", 
	  it->dn2 ? ((treebank_label)it->dn2->contents)->label : "NIL", 
	  it->df ? ((treebank_label)it->df->contents)->label : "NIL",
	  it->n ? item_a_hash(it):0);
}

void item_merge(item it1, item it2) {
  it1->inferences = list_nconc(it2->inferences, it1->inferences);
  it2->inferences = NULL;

  item_delete(it2);
}


/*

  Virtual unstopped        dn
  Virtual stopped          dn, en
  Non-virtual unstopped    dn
  Non-virtual stopped      dn
    pred root              df
    mod root               dn, type
    interior               dn, en, order
  
*/

unsigned int top_hash(node en, node dn) {
  return ((1*2 + node_virtual(en))*2 + (unsigned int)dn)*2 + (unsigned int)en;
}

unsigned int bottom_hash(int virtual, node dn) {
  return (1*2 + virtual)*2 + (unsigned int)dn;
}

unsigned int dn_hash(node dn) {
  return (unsigned int)dn;
}

unsigned int modroot_hash(node dn, int tree_type) {
  return tree_type*2 + (unsigned int)dn;
}

unsigned int item_a_hash(item it) {
  if (!code_equal(it->code, CODE_STOP))
    return bottom_hash(node_virtual(it->n), node_virtual(it->n) == LEFT ? it->dn1 : it->dn2);
  else if (node_parent(it->n) == NULL) {
    if (tree_type(it->n) == LEFTPRED || tree_type(it->n) == RIGHTPRED)
      return dn_hash(it->df);
    else
      return modroot_hash(it->dn1, tree_type(it->n));
  } else {
    return top_hash(it->n, node_virtual(it->n) == LEFT ? it->dn1 : it->dn2);
  }
}

unsigned int item_r_hash(item it) {
  return (unsigned int)(it->n) * 2 + (unsigned int)(it->dn1);
}

int match(node en, node dn) {
  if (en == NULL && dn == NULL)
    return 1;
  if (en == NULL || dn == NULL)
    return 0;
  return ((elem_label)en->contents)->label == dict_lookup(d, ((treebank_label)dn->contents)->label);
}

int item_compare(item it1, item it2) {
  return !(it1->n == it2->n && it1->dn1 == it2->dn1 && it1->dn2 == it2->dn2 && it1->df == it2->df && code_equal(it1->code, it2->code));
}

void seed_foot(chart c, node en, node dn) {
  int i, n;

  n = node_n_children(dn);

  if (n != 0 && match(en, dn))
    chart_add(c, new_item(en, dn, dn, dn, CODE_STOP, OP_NULL, NULL, NULL));

  for (i=0; i<n; i++)
    seed_foot(c, en, node_child(dn, i));
}

void seed_leaf(chart c, node en, node dn) {
  int i, n;

  n = node_n_children(dn);

  if (n == 0 && match(en, dn))
    chart_add(c, new_item(en, dn, dn, NULL, CODE_STOP, OP_NULL, NULL, NULL));

  for (i=0; i<n; i++)
    seed_leaf(c, en, node_child(dn, i));
}

int seed(chart c, node root, node tree) {
  int i;
  if (node_n_children(root) == 0) {
    if (node_type(root) == FOOT) {
      seed_foot(c, root, tree);
/*    } else if (node_type(root) == FRONTIER && node_empty(root)) {
      for (i=0; i<=n; i++)
	add(root, CODE_STOP, i, i, 0.0, ACTION_NODE, NULL, NULL);*/
    } else if (node_type(root) == FRONTIER) {
      seed_leaf(c, root, tree);
    }
  } else {
    for (i=0; i<node_n_children(root); i++)
      seed(c, node_child(root, i), tree);
  }

  return 0;
}


void add_axioms(chart c, node root) {
  list_node li;

  for (li = subgrammar; li != NULL; li = li->next)
    seed(c, tree_root(li->contents), root);

}


void select_subst(node cur) {
  int i, n;

  n = node_n_children(cur);
  if (node_type(cur) == SUBST)
    subst_nodes = new_list_node(cur, subst_nodes);
  for (i=0; i<n; i++)
    select_subst(node_child(cur, i));

}


void add_consequences(chart c, item it) {
  node eparent, esister, dparent, dsister;
  list_node li;
  item oit;
  int dir;

  eparent = node_parent(it->n);

  /* Unary concatenation */
  if (code_equal(it->code, CODE_STOP) && 
      eparent != NULL &&
      node_n_children(eparent) == 1) {

    /* Regular */
    if (!node_virtual(eparent)) {
      dparent = node_parent(it->dn1);
      if (dparent != NULL && match(eparent, dparent))
	chart_add(c, new_item(eparent, dparent, dparent, it->df, CODE_START, OP_NULL, it, NULL));
    }

    /* Virtual */
    else {
      chart_add(c, new_item(eparent, it->dn1, it->dn2, it->df, CODE_START, OP_NULL, it, NULL));
    }
  }


  /* Binary concatenation can happen in eight(!) different ways.

     Virtual, left, stopped     Regular, left child
     Virtual, left, unstopped   Regular, left modifier root
     Virtual, right, stopped    Regular, right child
     Virtual, right, unstopped  Regular, right modifier root

     search on dn1 or dn2       search on dn1 == dn2

     */

  /* Sister adjunction */

  /* Virtual, unstopped */
  /* Search for dn, code, tree_type */
  if (node_virtual(it->n) && !code_equal(it->code, CODE_STOP)) {
    dir = node_virtual(it->n);
    dsister = dir == LEFT ? node_sister(it->dn1, -1) : node_sister(it->dn2, 1);
    if (dsister != NULL) {
      for (li = hash_map_similar(c->a_index, modroot_hash(dsister, dir == LEFT ? LEFTMOD : RIGHTMOD)); li != NULL; li = li->next) {
	oit = ((struct hash_map_entry *)li->contents)->key;

	if (node_parent(oit->n) == NULL && code_equal(oit->code, CODE_STOP) &&
	    tree_type(oit->n) == (dir == LEFT ? LEFTMOD : RIGHTMOD) &&
	    oit->dn1 == dsister) {
	  if (dir == LEFT)
	    chart_add(c, new_item(it->n, dsister, it->dn2, it->df, 
				  compute_code(it->code, oit->n),
				  OP_SADJ, oit, it));
	  else
	    chart_add(c, new_item(it->n, it->dn1, dsister, it->df, 
				  compute_code(it->code, oit->n),
				  OP_SADJ, oit, it));
	}
      }
    }
  }

  /* Modifier root, stopped */
  /* Search for dn, code, virtual */
  if (!node_virtual(it->n) && eparent == NULL && 
      (tree_type(it->n) == LEFTMOD || tree_type(it->n) == RIGHTMOD) &&
      code_equal(it->code, CODE_STOP)) {

    /* We need two blocks because it so happens that LEFTMOD and RIGHTMOD
       are the same. */

    if (tree_type(it->n) == LEFTMOD) {
      dir = LEFT;
      dsister = node_sister(it->dn2, 1);
      if (dsister != NULL) {
	
	for (li = hash_map_similar(c->a_index, bottom_hash(dir, dsister)); li != NULL; li = li->next) {
	  oit = ((struct hash_map_entry *)li->contents)->key;
	  
	  if (node_label(oit->n) == left_wild &&
	      !code_equal(oit->code, CODE_STOP) &&
	      oit->dn1 == dsister) {
	    chart_add(c, new_item(oit->n, it->dn1, oit->dn2, 
				  oit->df, compute_code(oit->code, it->n),
				  OP_SADJ, it, oit));
	  }
	}
      }
    }
    
    if (tree_type(it->n) == RIGHTMOD) {
      dir = RIGHT;
      dsister = node_sister(it->dn1, -1);
      if (dsister != NULL) {
	
	for (li = hash_map_similar(c->a_index, bottom_hash(dir, dsister)); li != NULL; li = li->next) {
	  oit = ((struct hash_map_entry *)li->contents)->key;
	  
	  if (trace) {
	    printf ("  with   ");
	    item_print(stdout, oit);
	    printf ("\n");
	  }

	  if (node_label(oit->n) == right_wild &&
	      !code_equal(oit->code, CODE_STOP) &&
	      oit->dn2 == dsister) {
	    chart_add(c, new_item(oit->n, oit->dn1, it->dn2,
				  oit->df, compute_code(oit->code, it->n),
				  OP_SADJ, it, oit));
	  }
	}
      }
    }
    
  }
  
  /* Binary concatenation */

  /* Virtual, stopped */
  /* Search for dn, en, code */
  if (node_virtual(it->n) && code_equal(it->code, CODE_STOP) && node_n_children(eparent) == 2) {
    dir = node_virtual(it->n);
    dsister = dir == LEFT ? node_sister(it->dn1, -1) : node_sister(it->dn2, 1);
    esister = dir == LEFT ? node_sister(it->n, -1) : node_sister(it->n, 1);
    if (dsister != NULL) {
      for (li = hash_map_similar(c->a_index, top_hash(esister, dsister)); li != NULL; li = li->next) {
	oit = ((struct hash_map_entry *)li->contents)->key;

	if (code_equal(oit->code, CODE_STOP)) {
	  if (dir == LEFT && oit->n == esister && oit->dn1 == dsister)
	    chart_add(c, new_item(eparent, dsister, it->dn2, it->df ? it->df : oit->df,
				  CODE_START,
				  OP_NULL, oit, it));
	  else if (dir == RIGHT && oit->n == esister && oit->dn1 == dsister)
	    chart_add(c, new_item(eparent, it->dn1, dsister, it->df ? it->df : oit->df,
				  CODE_START,
				  OP_NULL, it, oit));
	}

      }

    }

  }

  /* Interior, stopped */
  /* Search for dn, en, code, virtual */
  if (!node_virtual(it->n) && code_equal(it->code, CODE_STOP) && 
      eparent != NULL && node_n_children(eparent) == 2) {
    
    dir = node_order(it->n) == 0 ? LEFT : RIGHT;
    dsister = dir == LEFT ? node_sister(it->dn2, 1) : node_sister(it->dn1, -1);
    esister = dir == LEFT ? node_sister(it->n, 1) : node_sister(it->n, -1);

    /*    if (dsister) {
      static char buf[100];
      treebank_label_to_string(dsister->contents, buf, sizeof(buf));
      printf(" dsister %s", buf);
    }
    if (esister) {
      static char buf[100];
      elem_label_to_string(esister->contents, buf, sizeof(buf));
      printf(" esister %s", buf);
      }*/
      

    if (dsister != NULL) {

      for (li = hash_map_similar(c->a_index, top_hash(esister, dsister)); li != NULL; li = li->next) {
	oit = ((struct hash_map_entry *)li->contents)->key;

	if (trace) {
	  printf ("  with   ");
	  item_print(stdout, oit);
	  printf ("\n");
	}

	  
	if (code_equal(oit->code, CODE_STOP)) {
	  if (dir == LEFT && oit->n == esister && oit->dn1 == dsister)
	    chart_add(c, new_item(eparent, it->dn1, oit->dn2,
				  it->df ? it->df : oit->df, CODE_START,
				  OP_NULL, it, oit));
	  else if (dir == RIGHT && oit->n == esister && oit->dn2 == dsister)
	    chart_add(c, new_item(eparent, oit->dn1, it->dn2,
				  it->df ? it->df : oit->df, CODE_START,
				  OP_NULL, oit, it));
	}
      }
    }
  }
  
  /* Predicative adjunction */

  /* Predicative root, stopped */
  /* search on dn */
  if (code_equal(it->code, CODE_STOP) && eparent == NULL && 
      (tree_type(it->n) == LEFTPRED || tree_type(it->n) == RIGHTPRED)) {

    for (li = hash_map_similar(c->a_index, bottom_hash(0, it->df)); li != NULL; li = li->next) {
      oit = ((struct hash_map_entry *)li->contents)->key;


      if (trace) {
	printf ("  with   ");
	item_print(stdout, oit);
	printf ("\n");
      }

      if (!code_equal(oit->code, CODE_STOP) && 
          !node_virtual(oit->n) &&
          oit->dn1 == it->df &&
          node_adjoinable(it->n, oit->n)) {
        chart_add(c, new_item(oit->n, it->dn1, it->dn2, oit->df, CODE_STOP,
			      OP_ADJ, it, oit));
      }
    }
  }

  /* Non-virtual, unstopped */
  /* search on df */
  if (!code_equal(it->code, CODE_STOP) && !node_virtual(it->n)) {

    for (li = hash_map_similar(c->a_index, dn_hash(it->dn1)); li != NULL; li = li->next) {
      oit = ((struct hash_map_entry *)li->contents)->key;
      if (trace) {
	printf ("  with   ");
	item_print(stdout, oit);
	printf ("\n");
      }

      if (code_equal(oit->code, CODE_STOP) &&
	  node_parent(oit->n) == NULL &&
	  (tree_type(oit->n) == LEFTPRED || tree_type(oit->n) == RIGHTPRED) &&
	  oit->df == it->dn1 &&
	  node_adjoinable(oit->n, it->n)) {
	chart_add(c, new_item(it->n, oit->dn1, oit->dn2, it->df, CODE_STOP,
			      OP_ADJ, oit, it));
      }
    }
  }

  /* Stop sister adjunction */
  if (!code_equal(it->code, CODE_STOP) &&
      (node_virtual(it->n) == RIGHT && match(node_sister(it->n, 1), node_sister(it->dn2, 1)) ||
       node_virtual(it->n) == LEFT && match(node_sister(it->n, -1), node_sister(it->dn1, -1))))
    chart_add(c, new_item(it->n, it->dn1, it->dn2, it->df, CODE_STOP,
			  OP_STOPSADJ, it, NULL));

  /* No adjunction */
  if (!code_equal(it->code, CODE_STOP) && !node_virtual(it->n))
    chart_add(c, new_item(it->n, it->dn1, it->dn2, it->df, CODE_STOP,
			  OP_NOADJ, it, NULL));

  /* Goal / Substitution */
  if (code_equal(it->code, CODE_STOP) && eparent == NULL && tree_type(it->n) == INITIAL) {

    if (node_parent(it->dn1) == NULL) {
      chart_add_goal(c, new_item(NULL, it->dn1, it->dn2, NULL, CODE_STOP,
				 OP_INITIAL, it, NULL));
    } else {
      for (li = subst_nodes; li != NULL; li = li->next)
	if (node_substitutable(it->n, li->contents))
	  chart_add(c, new_item(li->contents, it->dn1, it->dn2, it->df, CODE_STOP, OP_SUBST, it, NULL));
    }
  }
}

void compute_viterbi_helper(item it) {
  list_node li;
  struct inference *inf;
  double p;

  if (it->flag == 1)
    return;
  it->flag = 1;

  for (li = it->inferences; li != NULL; li = li->next) {
    inf = li->contents;
    
    if (inf->ant1) {
      compute_viterbi_helper(inf->ant1);
      if (inf->ant2) {
	compute_viterbi_helper(inf->ant2);
	p = inf->ant1->p_viterbi * inf->ant2->p_viterbi * inf->p;
      } else {
	p = inf->ant1->p_viterbi * inf->p;
      }
    } else {
      p = inf->p;
    }

    if (p > it->p_viterbi)
      it->p_viterbi = p;
  }
}

void compute_viterbi(chart c) {
  compute_viterbi_helper(c->goal);
}

void compute_inside_helper(item it) {
  list_node li;
  struct inference *inf;

  /*  if (it->flag == 1)
    return;
  it->flag = 1;

    for (li = it->inferences; li != NULL; li = li->next) {
    inf = li->contents;
    if (inf->ant1) {
      compute_inside_helper(inf->ant1);
      if (inf->ant2) {
	compute_inside_helper(inf->ant2);
      }
    }
    }*/

  if (trace) {
    item_print(stdout, it);
    printf ("p_inside = %g", it->p_inside);
  }
  for (li = it->inferences; li != NULL; li = li->next) {
    inf = li->contents;


    if (inf->ant1) {
      if (inf->ant2) {
	it->p_inside += inf->ant1->p_inside * inf->ant2->p_inside * inf->p;
	if (trace)
	  printf(" + %g * %g * %g", inf->ant1->p_inside * inf->ant2->p_inside * inf->p);
      } else {
	it->p_inside += inf->ant1->p_inside * inf->p;
	if (trace)
	  printf(" + %g * %g", inf->ant1->p_inside * inf->p);
      }
    } else {
      it->p_inside += inf->p;
	if (trace)
	  printf(" + %g", inf->p);
    }
  }
  if (trace)
    printf("\n");
}

void compute_inside(chart c) {
  int i;
  for (i=0; i<c->chart_size; i++)
    compute_inside_helper(c->items[i]);
  compute_inside_helper(c->goal);
}

/* doesn't work right: can't recurse until consequent has been totally computed */

void compute_outside_helper(item it) {
  list_node li;
  struct inference *inf;

  /*  if (it->flag == 2)
    return;
    it->flag = 2;*/

  for (li = it->inferences; li != NULL; li = li->next) {
    inf = li->contents;
    
    if (inf->ant1) {
      if (inf->ant2) {
	inf->ant1->p_outside += it->p_outside * inf->ant2->p_inside * inf->p;
	inf->ant2->p_outside += it->p_outside * inf->ant1->p_inside * inf->p;
	/*	compute_outside_helper(inf->ant2);*/
      } else {
	inf->ant1->p_outside += it->p_outside * inf->p;
      }
      /*      compute_outside_helper(inf->ant1);*/
    }
  }
}

void compute_outside(chart c) {
  int i;
  
  c->goal->p_outside = 1.0;
  compute_outside_helper(c->goal);
  for (i=c->chart_size-1; i>=0; i--)
    compute_outside_helper(c->items[i]);
}

void count_generic(FILE *fp, int *ks, double delta) {
  static char buf[100];
  /* copied from model.c */
  static char *names[10] = {
    "tree'", "t'", "w'",
    "tree", "t", "w",
    "type", "addr", "code",
    "NIL" };
  int i;
  fprintf(fp, "%lg", delta);
  for (i=0; i<10; i++) {
    dict_rlookup(d, ks[i], buf, sizeof(buf));
    fprintf(fp, " %s=%s", names[i], buf);
  }
  fprintf(fp, "\n");
}

void count_initial(FILE *fp, node r, double delta) {
  elem_label sr = node_contents(r);
  static int ks[10];

  ks[0] = sr->pge->d_index;
  ks[1] = sr->headtag;
  ks[2] = sr->headword;
  ks[3] = ks[4] = ks[5] = ks[6] = ks[7] = ks[8] = ks[9] = nil;

  count_generic(fp, ks, delta);
}

void count_subst(FILE *fp, node r, node s, double delta) {
  elem_label sr = node_contents(r), ss = node_contents(s);
  static int ks[10];

  ks[0] = sr->pge->d_index;
  ks[1] = sr->headtag;
  ks[2] = sr->headword;
  ks[3] = ss->pge->d_index;
  ks[4] = ss->headtag;
  ks[5] = ss->headword;
  ks[6] = ss->d_type;
  ks[7] = ss->d_addr;
  ks[8] = CODE_START.code;
  ks[9] = nil;

  count_generic(fp, ks, delta);
}

/* N.B. the left_to_right option must be turned off, 
   so we don't reproduce that part of p_adj() here */

void count_adj(FILE *fp, node r, node s, code_t code, double delta) {
  elem_label sr = node_contents(r), ss = node_contents(s);

  static int ks[10];
  prob cur;
  float p;

  if (sr == NULL) {
    if (ss->noadj || ss->type == FRONTIER)
      return;

    ks[0] = stop;
    ks[1] = stop;
    ks[2] = stop;
    ks[3] = ss->pge->d_index;
    ks[4] = ss->headtag;
    ks[5] = ss->headword;
    ks[6] = ss->d_type;
    ks[7] = ss->d_addr;
    ks[8] = code.code;
    ks[9] = nil;
  } else {
    ks[0] = sr->pge->d_index;
    ks[1] = sr->headtag;
    ks[2] = sr->headword;
    ks[3] = ss->pge->d_index;
    ks[4] = ss->headtag;
    ks[5] = ss->headword;
    ks[6] = ss->d_type;
    ks[7] = ss->d_addr;
    ks[8] = code.code;
    ks[9] = nil;
  }
  count_generic(fp, ks, delta);
}

void reestimate_helper(FILE *fp, item it, double total) {
  list_node li;
  int op;
  item ant1, ant2;
  double delta = 0.0, p;

  for (li = it->inferences; li != NULL; li = li->next) {
    op = ((struct inference *)li->contents)->op;
    ant1 = ((struct inference *)li->contents)->ant1;
    ant2 = ((struct inference *)li->contents)->ant2;
    p = ((struct inference *)li->contents)->p;

    if (ant1) {
      if (ant2) {
	delta = it->p_outside * ant1->p_inside * ant2->p_inside * p / total;
      } else {
	delta = it->p_outside * ant1->p_inside * p / total;
      }
    }

    if (delta > 0.0) {

      if (trace) {
	printf("p %g e %g f %g ", p, it->p_inside, it->p_outside);
	if (ant1)
	  printf("e1 %g f1 %g ", ant1->p_inside, ant1->p_outside);
	if (ant2)
	  printf("e2 %g f2 %g ", ant2->p_inside, ant2->p_outside);
      }

      switch (op) {
      case OP_INITIAL: count_initial(fp, ant1->n, delta); break;
      case OP_SUBST: count_subst(fp, ant1->n, it->n, delta); break;
      case OP_ADJ: 
      case OP_SADJ: count_adj(fp, ant1->n, ant2->n, ant2->code, delta); break;
      case OP_NOADJ:
      case OP_STOPSADJ: count_adj(fp, NULL, ant1->n, ant1->code, delta); break;
      case OP_NULL: 
	if (trace) {
	  printf("Null\n");
	}
	break;
      }
    }
  }
}

void reestimate(FILE *fp, chart c) {
  int i;
  for (i=0; i<c->chart_size; i++)
    reestimate_helper(fp, c->items[i], c->goal->p_inside);
  reestimate_helper(fp, c->goal, c->goal->p_inside);
}

void compute_atomic_helper(item it) {
  list_node li;
  struct inference *inf;

  for (li = it->inferences; li != NULL; li = li->next) {
    inf = li->contents;

    switch (inf->op) {
    case OP_INITIAL: inf->p = exp10(p_initial(inf->ant1->n)); break;
    case OP_SUBST: inf->p = exp10(p_subst(inf->ant1->n, it->n)); break;
    case OP_ADJ: 
    case OP_SADJ: inf->p = exp10(p_adj(inf->ant1->n, inf->ant2->n, inf->ant2->code)); break;
    case OP_NOADJ:
    case OP_STOPSADJ: inf->p = exp10(p_adj(NULL, inf->ant1->n, inf->ant1->code)); break;
    case OP_NULL: inf->p = 1.0; break;
    }
  }
}

void compute_atomic(chart c) {
  int i;
  for (i=0; i<c->chart_size; i++)
    compute_atomic_helper(c->items[i]);
  compute_atomic_helper(c->goal);
}

void topological_sort(chart c) {
  int i, j;
  item *items;
  list_node li;
  struct inference *inf;
  int axiom, done;

  items = malloc(c->chart_size * sizeof(item));

  j = 0;
  done = 0;
  while (!done) {
    done = 1;
    for (i=0; i<c->chart_size; i++) {
      if (c->items[i] != NULL) {
	axiom = 1;
	/* Look for undeleted antecedents */
	for (li = c->items[i]->inferences; li != NULL; li = li->next) {
	  inf = li->contents;
	  if (inf->ant1)
	    axiom = axiom && inf->ant1->flag;
	  if (inf->ant2)
	    axiom = axiom && inf->ant2->flag;
	}
	
	/* If none, we can append it to the list */
	if (axiom) {
	  done = 0;
	  items[j] = c->items[i];
	  c->items[i] = NULL;
	  items[j]->flag = 1; /* mark as deleted */
	  j++;
	}
      }
    }
  }

  if (j != c->chart_size)
    fprintf(stderr, "topological_sort(): graph is cyclic\n");
  else {
    free(c->items);
    c->items = items;
  }
}

/*** Interface to grammar.c ***/

float beam(int span) {
  return -12.0;
}

/*** Main program ***/

int main (int argc, char *argv[]) {
  chart c;
  item goal;
  node tree;
  FILE *fp;
  static char filename[FILENAME_MAX], events_filename[FILENAME_MAX];
  time_t start = time(NULL);
  list_node grammar, li;
  static struct word *w[MAX_WORDS];
  int i, n;
  int ch;
  struct word *w1;
  float ll = 0.0;

  strcpy(events_filename, "events");

  /* Process command-line options */
  while ((ch = getopt(argc, argv, "t:w:hp:e:C:")) != EOF)
    switch (ch) {
    case 't': tree_threshold = atoi(optarg); break;
    case 'w': word_threshold = atoi(optarg); break;
    case 'h': left_to_right = 0; break;
    case 'p': strncpy(prefix, optarg, FILENAME_MAX); break;
    case 'e': strncpy(events_filename, optarg, FILENAME_MAX); break;
    case 'C': C = atof(optarg); break;
    }

  debug = 0;

  model_init();
  unknown = dict_lookup(d, "*UNKNOWN*");
  nil = dict_lookup(d, "NIL");
  stop = dict_lookup(d, "*STOP*");
  left_wild = dict_lookup(d, "<");
  right_wild = dict_lookup(d, ">");

  words = new_lexicon((COMPARE)word_compare, (HASH)word_hash);
  sprintf(filename, "%s.lex", prefix);
  fp = fopen(filename, "r");
  lexicon_read(words, fp, (READ)read_word);
  fclose(fp);

  fp = fopen(events_filename, "w");

  if (stats)
    fprintf(stderr, "(%d s) Done initializing\n", time(NULL)-start);

  while ((tree = read_tree(stdin, (FROM_STRING)treebank_nonterm_from_string, (FROM_STRING)treebank_term_from_string)) != NULL) {
    
    n = tree_to_sent(tree, w, MAX_WORDS);
    if (stats)
      fprintf(stderr, "Read sentence:");
    
    for (i=0; i<n; i++) {
      w1 = lexicon_lookup(words, w[i]);
      /*    if (w1 != NULL && replace_tags)
	    word_merge(w[i], w1);*/
      if (lexicon_count(words, w[i]) < word_threshold)
	w[i]->w = unknown;
      
      if (stats) {
	fprintf(stderr, " ");
	write_word(stderr, w[i]);
      }
      
    }
    
    if (stats)
      fprintf(stderr, "\n");
    
    grammar = growTrees(w, n);
    subst_nodes = NULL;
    for (li = grammar; li != NULL; li = li->next)
      select_subst(tree_root(li->contents));

    c = new_chart();

    add_axioms(c, tree);
    chart_parse(c);

    if (stats)
      fprintf(stderr, "(%d s) Done parsing\n", time(NULL)-start);

    if (c->goal) {
      topological_sort(c);
      compute_atomic(c);
      compute_inside(c);
      compute_outside(c);
      reestimate(fp, c);
      if (stats)
	fprintf(stderr, "(%d s) Done reestimating\n", time(NULL)-start);
      fprintf(stdout, "%g\n", c->goal->p_inside);
    } else {
      if (stats)
	fprintf(stderr, "(%d s) warning: no parse\n", time(NULL)-start);
      fprintf(stdout, "0.0\n");
    }

    /*    tree_print(stdout, tree, (TO_STRING)treebank_label_to_string);*/

    fflush(fp);
    fflush(stdout);

    chart_delete(c);
    tree_delete(tree, (DELETE)treebank_label_delete);

    deleteTrees(grammar);
    list_delete(subst_nodes);

  }
  fclose(fp);
}

