/* Copyright (c) 2001 by David Chiang. All rights reserved.*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "treebank_label.h"

int print_tags = 0;
int print_index = 0;

#define MAX_LABELS 128
#define MAX_TAGS 32

treebank_label new_treebank_label(void) {
  treebank_label tl = malloc(sizeof(*tl));

  tl->s = NULL;
  tl->label = NULL;
  tl->labels = NULL;
  tl->tags = NULL;
  tl->n_labels = 0;
  tl->n_tags = 0;
  tl->coord_index = 0;
  tl->ref_index = 0;

  return tl;
}

void treebank_label_delete(treebank_label tl) {
  free (tl->s);
  if (tl->labels)
    free (tl->labels);
  free (tl->tags);

  free(tl);
}

int treebank_label_to_string(treebank_label tl, char *s, size_t n) {
  static char buf[40];
  int i;

  if (tl->label) {

    if (!tl->labels) {
      strncpy(s, tl->label, n);
    } else {
      strncpy(s, tl->labels[0], n);
      for (i=1; i<tl->n_labels; i++) {
	sprintf(buf, "|%s",tl->labels[i]);
	strncat(s, buf, n-strlen(tl->label));
      }
    }

    if (print_tags) {
      for (i=0; i<tl->n_tags; i++) {
	sprintf(buf, "-%s", tl->tags[i]);
	strncat(s, buf, n-strlen(tl->label));
      }
    }

    if (tl->coord_index > 0 && print_index) {
      sprintf(buf, "=%d", tl->coord_index);
      strncat(s, buf, n-strlen(tl->label));
    }
    if (tl->ref_index > 0 && print_index) {
      sprintf(buf, "-[%d]", tl->ref_index);
      strncat(s, buf, n-strlen(tl->label));
    }
  } else {
    *s = '\0';
  }
  return 0;
}

/* Treebank labels have the following general format:

   terminals: <label> | <empty-cat>(-<ref-index>)?

   where <empty-cat> is of the form \*[A-Z]+\*

   nonterminals: <label>(-<tag>)*(=<coord-index>)?(-<ref-index>)? 

   where <label> is either -NONE- or of the form [A-Z]+

*/

treebank_label treebank_term_from_string(const char *s) {
  treebank_label tl = new_treebank_label();
  static regex_t reg;
  static int initialized = 0;
  const int nmatch = 10;
  regmatch_t match[nmatch];

  if (!initialized) {
    regcomp(&reg, "^(\\*([A-Za-z]+\\*)?)-([0-9]+)$", REG_EXTENDED);
    initialized = 1;
  }

  tl->s = strdup(s);

  if (!strchr(s, '-')) { /* common case to bypass regex matching */
    tl->label = tl->s;
  } else if (regexec(&reg, s, nmatch, match, 0) == 0) {
    tl->label = tl->s+match[1].rm_so;
    tl->s[match[1].rm_eo] = '\0';
    tl->ref_index = atoi(s+match[3].rm_so);
  } else {
    tl->label = tl->s;
  }

  return tl;
}

treebank_label treebank_nonterm_from_string(const char *s) {
  treebank_label tl = new_treebank_label();
  int i, j, k;
  static regex_t reg_nonterm;
  static int initialized = 0;
  const int nmatch = 10;
  regmatch_t match[nmatch];
  static char *labels[MAX_LABELS], *tags[MAX_TAGS];

  if (!initialized) {
    /* We allow * in nonterminals so wildcards can be used in the 
       table files */
    regcomp(&reg_nonterm, "^([A-Za-z|*$]+)((-[A-Za-z]+)*)(=[0-9]+)?(-[0-9]+)?$", REG_EXTENDED);
/*    regcomp(&reg_tag, "^-([A-Z]+)", REG_EXTENDED);*/
    initialized = 1;
  }

  tl->s = strdup(s);

  if ((!strchr(s, '-') && !strchr(s, '=') && !strchr(s, '|')) ||  /* common case to bypass regex matching */ 
      !strcmp(s, "-NONE-")) {
    tl->n_labels = 1;
    tl->labels = malloc(1*sizeof(char *));
    tl->label = tl->labels[0] = tl->s;
  } else {
    if (regexec(&reg_nonterm, s, nmatch, match, 0) == 0) {
      /* Extract label(s) */
      
      i = 0;
      k = match[1].rm_so;
      for (j=match[1].rm_so; j<match[1].rm_eo; j++)
	if (s[j] == '|') {
	  tl->s[j] = '\0';
	  if (i<MAX_LABELS)
	    labels[i] = tl->s+k;
	  i++;
	  k = j+1;
	}
      tl->s[j] = '\0';
      if (i<MAX_LABELS)
	labels[i] = tl->s+k;
      i++;

      tl->n_labels = i;
      tl->labels = malloc(i*sizeof(char *));
      for (i=0; i<tl->n_labels; i++)
	tl->labels[i] = labels[i];
      tl->label = labels[0];

      /* Extract indices */
      
      if (match[4].rm_so >= 0)
	tl->coord_index = atoi(s+match[4].rm_so+1); /* +1 for = */
      if (match[5].rm_so >= 0)
	tl->ref_index = atoi(s+match[5].rm_so+1); /* +1 for - */
      
      /* Extract tags */
      
      if (match[2].rm_so >= 0) {
	i = 0;
	k = match[2].rm_so+1;
	for (j=match[2].rm_so+1; j<match[2].rm_eo; j++)
	  if (s[j] == '-') {
	    tl->s[j] = '\0';
	    if (i<MAX_TAGS)
	      tags[i] = tl->s+k;
	    i++;
	    k = j+1;
	  }
	if (i<MAX_TAGS)
	  tags[i] = tl->s+k;
	i++;

	tl->n_tags = i;
	tl->tags = malloc(i*sizeof(char *));
	for (i=0; i<tl->n_tags; i++)
	  tl->tags[i] = tags[i];
	
      }
    } else {
      tl->n_labels = 1;
      tl->labels = malloc(1*sizeof(char *));
      tl->label = tl->labels[0] = tl->s;
    }
  }
  
  return tl;
}

